/* SI 413 Fall 2011
 * ast.hpp
 * This is a C++ header file for the AST class hierarchy.
 */

#ifndef AST_HPP
#define AST_HPP

#include <fstream>
#include <string>
#include <vector>
#include <stack>
#include <set>
#include <map>
#include <cstdlib>
using namespace std;

#include "frame.hpp"

// This enum type gives codes to the different kinds of operators.
enum Oper {
  ADD, SUB,
  MUL, DIV,
  LT, GT, LE, GE, 
  EQ, NE,
  AND, OR, NOTOP
};

// Takes an Oper code and returns the corresponding operator string.
// This is used to generate the text in AST nodes, and for debugging.
const char* opstring(Oper op);

/* A Stmt is anything that can be evaluated at the top level such
 * as I/O, assignments, and control structures.
 * The last child of any statement is the next statement in sequence.
 */
class Stmt {
  private:
    Stmt* next;

  public:
    /* This static method is for building sequences of statements by the
     * parser. It takes two statements, and appends one after the other.
     * The returned value is a pointer to the new statement representing
     * the sequence.
     */
    static Stmt* append(Stmt* a, Stmt* b);

    /* Default constructor. The next statement pointer will be set to an
     * object of the NullStmt class. */
    Stmt();

    /* Using this constructor manually specifies the next statement in
     * sequence. */
    Stmt (Stmt* nextStmt) :next(nextStmt) { }

    /* This is the command that must be implemented everywhere to
     * execute this Stmt - that is, do whatever it is that this statement
     * says to do. */
    virtual void exec(Frame* f) = 0;

    /* This method just executes the current statement, and then
     * executes all the following statements, in sequence.
     * It can be overridden in sub-classes but usually won't need to be. */
    virtual void execAll(Frame* f) {
      exec(f);
      next->execAll(f);
    }

    /* Gets and sets the next statement in the sequence. */
    Stmt* getNext() { return next; }
    void setNext(Stmt* n) { next = n; }

    /* Just indicates whether this is an instance of the NullStmt class. */
    virtual bool isNull() { return false; }
};

/* Every AST node that is not a Stmt is an Exp.
 * These represent actual computations that return something
 */
class Exp {
 public:
  /* This is the method that must be overridden by all subclasses.
   * It should perform the computation specified by this node, and
   * return the resulting value that gets computed. */
  virtual void eval(Frame* f) = 0;
};

/* An identifier, i.e. variable or function name. */
class Id :public Exp {
  private:
    string val;

  public:
    // Constructor from a C-style string
    Id(const char* v) :val(v) { }

    // Returns a reference to the stored string value.
    string& getVal() { return val; }

    void eval(Frame* f);
};

/* A literal number in the program. */
class Num :public Exp {
  private:
    int val;

  public:
    Num(int v) :val(v) { }

    // To evaluate, just return the number!
    void eval(Frame* f);
};

/* A literal boolean value like "true" or "false" */
class BoolExp :public Exp {
  private:
    bool val;

  public:
    BoolExp(bool v) :val(v) { }

    void eval(Frame* f);
};

/* This class represents a binary operation. That is, an operation with
 * exactly two arguments, such as +, *, /, and, <=, etc. */
class BinOp :public Exp {
  protected:
    Exp* args[2];
    Oper op;

  public:
    BinOp(Exp* l, Oper o, Exp* r) :op(o) 
    { args[0] = l; args[1] = r; }
};

/* A binary opration for arithmetic, like + or *. */
class ArithOp :public BinOp {
  public:
    ArithOp(Exp* l, Oper o, Exp* r) :BinOp(l,o,r) { }

    void eval(Frame* f);
};

/* A binary operation for comparison, like < or !=. */
class CompOp :public BinOp {
  public:
    CompOp(Exp* l, Oper o, Exp* r) :BinOp(l,o,r) { }

    void eval(Frame* f);
};

/* A binary operation for boolean logic, like "and". */
class BoolOp :public BinOp {
  public:
    BoolOp(Exp* l, Oper o, Exp* r) :BinOp(l,o,r) { }

    void eval(Frame* f);
};

/* This class represents a unary operation. That is, an operation with
 * just one argument, such as "not" and -. */
class UnarOp :public Exp {
  protected:
    Oper op;
    Exp* arg;

  public:
    UnarOp(Oper o, Exp* l) :op(o), arg(l) { }
};

/* This class represents a unary negation operation. */
class NegOp :public UnarOp {
  public:
    NegOp(Exp* l) :UnarOp(SUB,l) { }

    void eval(Frame* f);
};

/* This class represents a unary "not" operation. */
class NotOp :public UnarOp {
  public:
    NotOp(Exp* l) :UnarOp(NOTOP,l) { }

    void eval(Frame* f);
};

/* This is a statement for a block of code, i.e., code enclosed
 * in curly braces { and }.
 */
class Block :public Stmt {
  private:
    Stmt* body;

  public:
    Block(Stmt* b) :body(b) { }

    Stmt* getBody() { return body; }

    void exec(Frame* f);
};

/* This class is necessary to terminate a sequence of statements. */
class NullStmt :public Stmt {
  public:
    NullStmt() :Stmt(NULL) { }
    
    bool isNull() { return true; }

    // Nothing to execute!
    void exec(Frame* f) { }
    void execAll(Frame* f) { }
};

/* This class is for plain "if" statements without an else clause. */
class IfStmt :public Stmt {
  private:
    Exp* clause;
    Stmt* ifblock;

  public:
    IfStmt(Exp* e, Stmt* s) :clause(e), ifblock(s) { }

    void exec(Frame* f);
};

/* This class is for if-else statements. */
class IfElse :public Stmt {
  private:
    Exp* clause;
    Stmt* ifblock;
    Stmt* elseblock;
   
  public:
    IfElse(Exp* e, Stmt* ib, Stmt* eb)
      :clause(e), ifblock(ib), elseblock(eb)
    { }

    void exec(Frame* f);
};

/* Class for while(...) {...} statements. */
class WhileStmt :public Stmt {
  private:
    Exp* clause;
    Stmt* block;
   
  public:
    WhileStmt(Exp* c, Stmt* b) :clause(c), block(b) { }

    void exec(Frame* f);
};

/* An assignment statement. This represents a RE-binding in the symbol table. */
class Asn :public Stmt {
  protected:
    Id* lhs;
    Exp* rhs;
   
    /* Returns the Id that we are assigning to. */
    Id* getLhs() { return lhs; }
    
  public:
    Asn(Id* l, Exp* r) :lhs(l), rhs(r) { }

    void exec(Frame* f);
};

/* A "new" statement is a subclass of assignment because it
 * has the same functionality as assignment, plus the additional
 * functionality that a new binding is created. */
class NewStmt :public Asn {
  public:
    NewStmt(Id* l, Exp* r) :Asn(l,r) { }

    void exec(Frame* f);
};

/* A write statement. */
class Write :public Stmt {
  private:
    Exp* val;

  public:
    Write(Exp* v) :val(v) { }

    // Executes this statement by writing the value of the expression to
    // standard out.
    void exec(Frame* f);
};

/* A read statement. */
class Read :public Stmt {
  private:
    Id* var;
   
  public:
    Read(Id* v) :var(v) { }

    void exec(Frame* f);
};

/* A lambda expression consists of a parameter name and a body. */
class Lambda :public Exp {
  private:
    Id* var;
    Block* body;

  public:
    Lambda(Id* v, Block* b) :var(v), body(b) { }

    // These getter methods are necessary to support actually calling
    // the lambda sometime after it gets created.
    string& getVar() { return var->getVal(); }
    Block* getBody() { return body; }

    void eval(Frame* f);
};

/* A function call consists of the function name, and the actual argument.
 * Note that all functions are unary. */
class Funcall :public Exp {
  private:
    Exp* funexp;
    Exp* arg;
  
  public:
    Funcall(Exp* f, Exp* a) :funexp(f), arg(a) { }

    void eval(Frame* f);
};

/* This class represents a do-nothing statement that only has an expression. */
class ExpStmt :public Stmt  {
  private:
    Exp* body;

  public:
    ExpStmt(Exp* b) :body(b) { }

    void exec(Frame* f);
};

#endif //AST_HPP
